/****************************************************************
 build-tree.C
 bmajoros@tigr.org

 This program builds an unpruned decision tree and stores it in a 
 file.  Other programs can be used to prune the tree and apply it
 to a data set.

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/

#include <iostream>
#include <stdlib.h>
#include <fstream>
#include "tigr++/TigrExceptions.H"
#include "tigr++/TigrString.H"
#include "tigr++/TigrStack.H"
#include "tigr++/TigrStacktrace.H"
#include "tigr++/TigrTime.H"
#include "tigr++/TigrCommandLine.H"
#include "tigr++/TigrBitSet.H"
#include "../NamesFile.H"
#include "../TrainingSet.H"
#include "../Schema.H"
#include "TreeNode.H"
#include "../Entropy.H"
#include "../AttrRangeTable.H"
#include "../AttrValuePair.H"
#include "../AttributeIterator.H"
#include "Classifier.H"

const int DEFAULT_GRANULARITY=10;

class Application
{
  float pruneThreshold;
  TigrString filestem;
  NamesFile namesFile;
  TrainingSet *trainingSet;
  Schema *schema;
  TreeNode *root;
  AttrRangeTable *rangeTable;
  Classifier *classifier;
  bool useGainRatio;
  int granularity;

  void buildTree();
  TigrBitSet *buildUniversalSet();
  Distribution *computeDistribution(TigrBitSet &);
  void computeDistribution(TigrBitSet &set,
			   Distribution &distribution);
  TreeNode *process(TempNode *);
  void replace(TempNode *,TreeNode *);
  TreeNode *computePredicate(TigrBitSet &,Distribution &);
  float computeDeltaH(AttrValuePair,TigrBitSet &,
		      Distribution &,TigrBitSet &leftSet,
		      TigrBitSet &rightSet,
		      Distribution &leftDist,
		      Distribution &rightDist);
  void filterPredicate(AttrValuePair,TigrBitSet &set,
		       TigrBitSet &leftSet,
		       TigrBitSet &rightSet);
  void filterDiscrete(AttrValuePair,TigrBitSet &set,
		       TigrBitSet &leftSet,
		       TigrBitSet &rightSet);
  void filterContinuous(AttrValuePair,TigrBitSet &set,
		       TigrBitSet &leftSet,
		       TigrBitSet &rightSet);
  bool evalDiscrete(AttrValuePair,Case &);
  bool evalContinuous(AttrValuePair,Case &);
  void generateOutput(const TigrString &filestem);
  void dumpTree(ostream &);
  void printSubtree(TreeNode *,int numTabs,ostream &);
  void runTestCases(TigrString &filename);
  void runTestCases(TrainingSet *);
public:
  Application(TigrString filestem,bool useGainRatio,int granularity);
  void go();
};


int main(int argc,char *argv[])
{
  TigrCommandLine commandLine(argc,argv,"rg:");
  if(commandLine.numArgs()!=1)
    {
      TigrString msg=TigrString("\n\
build-tree [-r] [-g granularity] <filestem>\n\
    where -r = use gain ratio\n\
          -g = use the given number of intervals when discretizing\n\
               continuous parameters (default is ")+DEFAULT_GRANULARITY+")\n\
";
      cerr << msg << endl;
      return 1;
    }
  bool useGainRatio=commandLine.option('r');
  int granularity=(commandLine.option('g') ? 
		   commandLine.optParm('g').asInt() : 
		   DEFAULT_GRANULARITY);

  try
    {
      TigrTime stopwatch;
      stopwatch.startCounting();

      Application app(commandLine.arg(0).c_str(),useGainRatio,granularity);
      app.go();

      stopwatch.stopCounting();
      cout << stopwatch.elapsedTime() << endl;
    }
  catch(RootException &e)
    {
      cerr << e.getMessage() << endl;
      return 1;
    }
  catch(string s)
    {
      cout<<"exception: "<<s<<endl;
    }
  catch(const char *p)
    {
      cout<<"exception: "<<p<<endl;
    }
  catch(...)
    {
      cout<<"unknown exception caught in main()"<<endl;
    }

  return 0;
}



Application::Application(TigrString filestem,bool useGainRatio,int granularity)
  : pruneThreshold(0.0), useGainRatio(useGainRatio),
    filestem(filestem), trainingSet(NULL), granularity(granularity)
{
  // ctor
}



void Application::go()
{
  TigrString namesFilename=filestem+(TigrString) ".names";
  TigrString trainFilename=filestem+(TigrString) ".data";
  TigrString testFilename=filestem+(TigrString) ".test";

  cerr<<"processing names file..."<<endl;
  schema=namesFile.processFile(namesFilename);

  cerr<<"loading training set..."<<endl;
  trainingSet=new TrainingSet(*schema,trainFilename);

  cerr<<"building attribute range table..."<<endl;
  rangeTable=new AttrRangeTable(*trainingSet,granularity);

  cerr<<"inducing tree..."<<endl;
  buildTree();

  cerr<<"generating output..."<<endl;
  generateOutput(filestem);
}



void Application::buildTree()
{
  // Allocate a stack
  TigrStack<TreeNode*> stack;

  // Push the root onto the stack
  TigrBitSet *universalSet=buildUniversalSet();
  Distribution *distribution=
    computeDistribution(*universalSet);
  stack.push(new TempNode(universalSet,distribution,NULL));
  
  // Continue processing until the stack is empty
  while(!stack.isEmpty())
    {
      // Get the node on top of the stack
      TempNode *tempNode=(TempNode *) stack.pop();

      // Replace it with a leaf or an internal node
      TreeNode *replacement=process(tempNode);
      replace(tempNode,replacement);
      delete tempNode;

      // Push children (if any) onto the stack
      if(replacement->getType()==PREDICATE_NODE)
	{
	  PredicateNode *pred=(PredicateNode *) replacement;
	  stack.push(pred->getLeftChild());
	  stack.push(pred->getRightChild());
	}
    }
}



TigrBitSet *Application::buildUniversalSet()
{
  int n=trainingSet->numCases();
  TigrBitSet *set=new TigrBitSet(n);
  int i;
  for(i=0 ; i<n ; ++i)
    set->addMember(i);
  return set;
}



void Application::replace(TempNode *oldNode,TreeNode *newNode)
{
  TreeNode *parent=oldNode->getParent();
  if(!parent)
    {
      root=newNode;
      return;
    }

  PredicateNode *pred=(PredicateNode *) parent;
  if(pred->getLeftChild()==oldNode)
    pred->attachLeftChild(newNode);
  else
    pred->attachRightChild(newNode);
}



Distribution *Application::computeDistribution(TigrBitSet &set)
{
  Distribution &distribution=
    *new Distribution(schema->getNumCategories());
  computeDistribution(set,distribution);

  return &distribution;
}



void Application::computeDistribution(TigrBitSet &set,
				      Distribution &distribution)
{
  distribution.setAllTo(0.0);

  int n=set.getMaxSize();
  int cardinality=0;
  for(int i=0 ; i<n ; ++i)
    if(set.isMember(i))
      {
	TrainingCase &c=trainingSet->getCase(i);
	int category=c.getCategory();
	distribution[category]+=1;
	++cardinality;
      }

  int numCat=schema->getNumCategories();
  if(cardinality>0)
    for(int i=0 ; i<numCat ; ++i)
      distribution[i]/=cardinality;
}



TreeNode *Application::process(TempNode *tempNode)
{
  TigrBitSet *set=tempNode->getSet();
  Distribution &distribution=*tempNode->getDistribution();
  float H=entropy(distribution);
  if(H<pruneThreshold || set->cardinality()==0)
    {
      // ### SHOULD ALSO STORE A PROBABILITY IN THE LEAF NODE...
      LeafNode *ln=
	new LeafNode(distribution.mostPopularCategory());
      return ln;
    }

  TreeNode *pred=computePredicate(*set,distribution);
  if(!pred)
    return new LeafNode(distribution.mostPopularCategory());

  return pred;
}



TreeNode *Application::computePredicate(TigrBitSet &set,
				  Distribution &distribution)
{
  // First, some initialization
  AttrValuePair bestPair(0,0);
  float bestDeltaH=-1.0;
  float originalH=entropy(distribution);

  int numCases=set.getMaxSize();
  TigrBitSet leftSet(numCases), rightSet(numCases);
  TigrBitSet bestLeftSet(numCases), bestRightSet(numCases);

  int numCategories=schema->getNumCategories();
  Distribution leftDist(numCategories);
  Distribution rightDist(numCategories);
  Distribution bestLeftDist(numCategories);
  Distribution bestRightDist(numCategories);

  // Iterate through the attributes-value pairs,
  // keeping track of the best one (the one that
  // reduces entropy the most)
  AttributeIterator iterator(*rangeTable);
  while(iterator.hasMoreElements())
    {
      AttrValuePair pair=iterator.currentElement();
      float deltaH=computeDeltaH(pair,set,distribution,
				 leftSet,rightSet,
				 leftDist,rightDist);
      if(deltaH>bestDeltaH)
	{
	  bestPair=pair;
	  bestDeltaH=deltaH;
	  bestLeftSet=leftSet;
	  bestRightSet=rightSet;
	  bestLeftDist=leftDist;
	  bestRightDist=rightDist;
	}
      iterator.advance();

      leftSet.purge();
      rightSet.purge();
    }

  if(bestDeltaH==0.0) return NULL;

  // Create a new predicate node for this attribute-
  // value pair
  PredicateNode *pNode=
    new PredicateNode(bestPair.getAttribute(),
		      bestPair.getFloatValue(),
		      bestDeltaH,
		      distribution.mostPopularCategory(),
		      originalH);

  // Attach TempNodes for the left and right children
  TigrBitSet *pLeftSet=new TigrBitSet(numCases);
  TigrBitSet *pRightSet=new TigrBitSet(numCases);
  *pLeftSet=bestLeftSet;
  *pRightSet=bestRightSet;
  Distribution *pLeftDist=new Distribution(bestLeftDist);
  Distribution *pRightDist=new Distribution(bestRightDist);
  TempNode *leftChild=
    new TempNode(pLeftSet,pLeftDist,pNode);
  TempNode *rightChild=
    new TempNode(pRightSet,pRightDist,pNode);
  pNode->attachLeftChild(leftChild);
  pNode->attachRightChild(rightChild);

  // Return the new predicate node
  return pNode;
}



float Application::computeDeltaH(AttrValuePair pair,
				 TigrBitSet &set,
				 Distribution &distribution,
				 TigrBitSet &leftSet,
				 TigrBitSet &rightSet,
				 Distribution &leftDist,
				 Distribution &rightDist)
{
  /*
    This method computes the change in entropy as a
    result of going down through a predicate node
   */

  float initH=entropy(distribution);
  
  filterPredicate(pair,set,leftSet,rightSet);

  computeDistribution(leftSet,leftDist);
  computeDistribution(rightSet,rightDist);

  float leftEntropy=entropy(leftDist);
  float rightEntropy=entropy(rightDist);

  int leftCardinality=leftSet.cardinality();
  int rightCardinality=rightSet.cardinality();
  
  int numCases=set.cardinality();
  float leftRatio=leftCardinality/(float)numCases;
  float rightRatio=rightCardinality/(float)numCases;
  float finalH=leftRatio*leftEntropy + rightRatio*rightEntropy;
  float deltaH=initH-finalH;

  if(useGainRatio)
    {
      /* 
	 Gain ratio = the reduction in classification entropy divided
	 by the attribute entropy
      */

      float pL=leftCardinality/(float)numCases;
      float pR=rightCardinality/(float)numCases;
      float attributeEntropy=-(pL*binaryLog(pL)+pR*binaryLog(pR));
      deltaH/=attributeEntropy; // <== gain ratio
    }

  return deltaH;
}



void Application::filterPredicate(AttrValuePair pair,
				  TigrBitSet &set,
				  TigrBitSet &leftSet,
				  TigrBitSet &rightSet)
{
  /*
    This method filters the given set by testing
    it against the predicate represented by the
    attribute-value pair.  Those cases passing the test
    go left, and all others go right.
   */

  switch((*schema)[pair.getAttribute()]->getType())
    {
    case DISCRETE:
      filterDiscrete(pair,set,leftSet,rightSet);
      break;
    case CONTINUOUS:
      filterContinuous(pair,set,leftSet,rightSet);
      break;
    }
}



void Application::filterDiscrete(AttrValuePair pair,
				 TigrBitSet &set,
				 TigrBitSet &leftSet,
				 TigrBitSet &rightSet)
{
  /*
    This method does the filtering (described above)
    for discrete attributes only
   */

  int i, numCases=set.getMaxSize();
  for(i=0 ; i<numCases ; ++i)
    if(set.isMember(i))
      {
	TrainingCase &c=trainingSet->getCase(i);
	if(evalDiscrete(pair,c))
	  leftSet.addMember(i);
	else
	  rightSet.addMember(i);
      }
}



void Application::filterContinuous(AttrValuePair pair,
				   TigrBitSet &set,
				   TigrBitSet &leftSet,
				   TigrBitSet &rightSet)
{
  /*
    This method does the filtering (described above)
    for continuous attributes only
   */

  int i, numCases=set.getMaxSize();
  for(i=0 ; i<numCases ; ++i)
    if(set.isMember(i))
      {
	TrainingCase &c=trainingSet->getCase(i);
	if(evalContinuous(pair,c))
	  leftSet.addMember(i);
	else
	  rightSet.addMember(i);
      }
}



bool Application::evalDiscrete(AttrValuePair pair,
			       Case &c)
{
  int attribute=pair.getAttribute();
  int value=pair.getIntValue();

  return c.getDiscrete(attribute)==value;
}



bool Application::evalContinuous(AttrValuePair pair,
				Case &c)
{
  int attribute=pair.getAttribute();
  float value=pair.getFloatValue();

  return c.getContinuous(attribute)<=value;
}



void Application::generateOutput(const TigrString &filestem)
{
  // First, write out human-readable tree:
  ofstream os((filestem+".tree").c_str());

  os << "\nDecision tree:" << endl;
  dumpTree(os);
  os << endl;

  // Now write out machine-readable tree:
  TigrString filename=filestem+".unpruned";
  DecisionTree tree(root);
  tree.save(filename.c_str());
}



void Application::dumpTree(ostream &os)
{
  printSubtree(root,0,os);
}



void Application::printSubtree(TreeNode *node,int numTabs,
			       ostream &os)
{
  int i;
  for(i=0 ; i<numTabs ; ++i)
    os << "|   ";

  switch(node->getType())
    {
    case LEAF_NODE:
      {
	LeafNode *leafNode=(LeafNode *) node;
	os << "category=" << leafNode->getCategory() << endl;
      }
      break;
    case PREDICATE_NODE:
      {
	PredicateNode *predicateNode=(PredicateNode *) node;
	int attribute=predicateNode->getAttribute();
	TigrString &name=namesFile.getName(attribute);
	os << name;
	switch((*schema)[attribute]->getType())
	  {
	  case DISCRETE:
	    os << "=" << predicateNode->getIntValue() << ":"
	       << endl;
	    break;
	  case CONTINUOUS:
	    os << "<=" << predicateNode->getFloatValue()
	       << ":" << endl;
	    break;
	  }
	printSubtree(predicateNode->getLeftChild(),
		     numTabs+1,os);
	printSubtree(predicateNode->getRightChild(),
		     numTabs+1,os);
      }
      break;
    case TEMP_NODE:
      throw TigrStacktrace("This shouldn't happen");
    }
}



void Application::runTestCases(TigrString &filename)
{
  ifstream is(filename.c_str());
  if(!is.good()) return; // no test cases

  TrainingSet *testSet=new TrainingSet(*schema,filename);

  runTestCases(testSet);
}



void Application::runTestCases(TrainingSet *testSet)
{  
  int numCorrect=0;
  int i, n=testSet->numCases();
  for(i=0 ; i<n ; ++i)
    {
      TrainingCase &c=testSet->getCase(i);
      int actual=c.getCategory();
      int predicted=classifier->classify(c);
      if(actual==predicted)
	++numCorrect;
    }

  float percentCorrect=100.0*numCorrect/(float)n;
  cout << "\t" << numCorrect 
       << " correct out of " << n 
       << " (" << percentCorrect << "%)" << endl;
}





