/****************************************************************
 prune-by-index.C
 bmajoros@tigr.org

 This program loads an unpruned tree generated by build-tree
 and prunes it to the specified size.  It then tests the resulting
 tree on the test set and reports results.  The pruned tree is
 also saved in a different file from the unpruned tree.

 The pruning threshold is specified by index.  The entropies of
 all nodes are sorted in increasing order, and adjacent pairs are
 averaged.  The index specified on the command line refers to the
 associated average, which is then used as the pruning threshold.

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/

#include <iostream>
#include <stdlib.h>
#include <fstream>
#include "tigr++/TigrExceptions.H"
#include "tigr++/TigrString.H"
#include "tigr++/TigrCommandLine.H"
#include "tigr++/TigrStack.H"
#include "tigr++/TigrTime.H"
#include "tigr++/TigrVector.H"
#include "tigr++/TigrStacktrace.H"
#include "../NamesFile.H"
#include "../TrainingSet.H"
#include "../Schema.H"
#include "TreeNode.H"
#include "../Entropy.H"
#include "../AttrRangeTable.H"
#include "../AttrValuePair.H"
#include "../AttributeIterator.H"
#include "Classifier.H"

class Application
{
  float pruneThreshold, percentCorrect;
  TigrString filestem;
  NamesFile namesFile;
  Schema *schema;
  TreeNode *root;
  Classifier *classifier;
  int pruneIndex;
  TigrVector<float> thresholds;
  bool timeSeries;
  TrainingSet *trainingSet, *testSet;

  void getPruneThresholds(DecisionTree &);
  int countCases(TigrString &filename);
  void generateOutput(const TigrString &filestem);
  void dumpTree(ostream &);
  void printSubtree(TreeNode *,int numTabs,ostream &);
  void runTestCases(TrainingSet *);
  TreeNode *prune(TreeNode *);
  void baseline();
  TigrBitSet *buildUniversalSet(TrainingSet &);
  Distribution *computeDistribution(TigrBitSet &,TrainingSet &);
  void computeDistribution(TigrBitSet &set,
			   Distribution &distribution,
			   TrainingSet &);
public:
  Application(TigrString filestem,int pruneIndex,bool timeSeries);
  void go();
};


void usage()
{
  cerr << "\
prune-by-index [-tc] <filestem> <prune-index>\n\
    -t = score categories relatively (for [t]ime-series prediction)\n\
    -c = [c]ount number of possible prune points\n\
"
       << endl;  
}


int main(int argc,char *argv[])
{
  TigrCommandLine cmd(argc,argv,"tc");
  if(cmd.numArgs()!=2 && !cmd.option('c')) 
    {
      usage();
      return -1;
    }

  try
    {
      TigrTime stopwatch;
      stopwatch.startCounting();

      bool timeSeries=cmd.option('t');
      int pruneIndex=(cmd.option('c') ? -1 : cmd.arg(1).asInt());
      Application app(cmd.arg(0),pruneIndex,timeSeries);
      app.go();

      stopwatch.stopCounting();
      cout << stopwatch.elapsedTime() << endl;
    }
  catch(RootException &e)
    {
      cout << e.getMessage() << endl;
      return -1;
    }
  catch(const string &msg)
    {
      cout << msg.c_str() << endl;
      return -1;
    }
  catch(const char *msg)
    {
      cout << msg << endl;
      return -1;
    }
  catch(...)
    {
      cout << "unknown exception caught in main()" << endl;
      return -1;
    }

  return 0;
}



Application::Application(TigrString filestem,int pruneIndex,
			 bool timeSeries)
  : pruneIndex(pruneIndex), filestem(filestem), 
    timeSeries(timeSeries)
{
  // ctor
}



void Application::go()
{
  // For filenames
  TigrString namesFilename=filestem+".names";
  TigrString trainFilename=filestem+".data";
  TigrString testFilename=filestem+".test";
  TigrString unpruned=filestem+".unpruned";

  // Load schema
  schema=namesFile.processFile(namesFilename);

  // Load training & test sets
  trainingSet=new TrainingSet(*schema,trainFilename);
  testSet=new TrainingSet(*schema,testFilename);

  // Load unpruned tree
  DecisionTree tree(NULL);
  tree.load(unpruned);

  // Get pruning threshold
  getPruneThresholds(tree);
  if(pruneIndex<0 || pruneIndex>thresholds.size())
    throw TigrString("Prune index must be between 0 and ")+thresholds.size();
  pruneThreshold=thresholds[pruneIndex-1];
  cout << "pruning with threshold " << pruneIndex << " out of "
       << thresholds.size() << " (" << pruneThreshold << ")" << endl;

  // Prune tree
  root=tree.getRoot();
  root=prune(root);

  DecisionTree newTree(root);
  classifier=new Classifier(newTree,*schema);

  // Store pruned tree in file
  generateOutput(filestem);

  // Evaluate the accuracy of the pruned tree
  if(testSet->numCases()>0)
    {
      cout << "Performance on test set:";
      runTestCases(testSet);
    }

  // Compute baseline accuracy for comparison
  baseline();
}



void Application::baseline()
{
  TigrBitSet *universalSetTrain=buildUniversalSet(*trainingSet);
  Distribution &trainingDistr=
    *computeDistribution(*universalSetTrain,*trainingSet);

  int majority=trainingDistr.mostPopularCategory();

  TigrBitSet *universalSetTest=buildUniversalSet(*testSet);
  Distribution &testDistr=
    *computeDistribution(*universalSetTest,*testSet);

  int numCategories=schema->getNumCategories();
  int majorityAccur=int(100*testDistr[majority]+0.5);
  int uniformAccur=int(100.0/numCategories+0.5);
  float nonuniformP=0.0;
  for(int i=0 ; i<numCategories ; ++i)
    nonuniformP+=trainingDistr[i]*testDistr[i];
  int nonuniformAccur=int(100*nonuniformP+0.5);
  
  TigrString majorityName=(*schema)[majority]->getName();
  cout<<"\tuniform random guess : \t\t\t"<<uniformAccur
      <<"% accuracy"<<endl;
  cout<<"\tnonuniform random guess : \t\t"<<nonuniformAccur
      <<"% accuracy"<<endl;
  cout<<"\tmajority rule (always predict "<<majorityName<<") : \t" 
      <<majorityAccur <<"% accuracy"<<endl;
  cout << "\tdecision tree : \t\t\t"<<int(0.5+percentCorrect)
       <<"% accuracy"<<endl;
}



TigrBitSet *Application::buildUniversalSet(TrainingSet &cases)
{
  int n=cases.numCases();
  TigrBitSet *set=new TigrBitSet(n);
  int i;
  for(i=0 ; i<n ; ++i)
    set->addMember(i);
  return set;
}



Distribution *Application::computeDistribution(TigrBitSet &set,
					       TrainingSet &trainingSet)
{
  Distribution &distribution=
    *new Distribution(schema->getNumCategories());
  computeDistribution(set,distribution,trainingSet);

  return &distribution;
}



void Application::computeDistribution(TigrBitSet &set,
				      Distribution &distribution,
				      TrainingSet &trainingSet)
{
  distribution.setAllTo(0.0);

  int n=set.getMaxSize();
  int cardinality=0;
  for(int i=0 ; i<n ; ++i)
    if(set.isMember(i))
      {
	//	cout<<n<<" : "<<i<<" -> "<<&trainingSet->getCase(i)<<endl;//###
	TrainingCase &c=trainingSet.getCase(i);
	int category=c.getCategory();
	distribution[category]+=1;
	++cardinality;
      }

  int numCat=schema->getNumCategories();
  if(cardinality>0)
    for(int i=0 ; i<numCat ; ++i)
      distribution[i]/=cardinality;
}



int Application::countCases(TigrString &filename)
{
  ifstream is(filename.c_str());
  if(!is.good())
    throw FileErrorException(filename.c_str(),"Cannot open file");
  int n=0;
  while(!is.eof())
    {
      TigrString line;
      line.getline(is);
      if(line.isEmpty())
	break;
      ++n;
    }

  return n;
}



void Application::generateOutput(const TigrString &filestem)
{
  TigrString pruned=filestem+".pruned";
  ofstream os(pruned.c_str());

  os << "\nDecision tree:" << endl;
  dumpTree(os);
  os << endl;
}



void Application::dumpTree(ostream &os)
{
  printSubtree(root,0,os);
}



void Application::printSubtree(TreeNode *node,int numTabs,
			       ostream &os)
{
  int i;
  for(i=0 ; i<numTabs ; ++i)
    os << "|   ";

  switch(node->getType())
    {
    case LEAF_NODE:
      {
	LeafNode *leafNode=(LeafNode *) node;
	os << "category=" << leafNode->getCategory() << endl;
      }
      break;
    case PREDICATE_NODE:
      {
	PredicateNode *predicateNode=(PredicateNode *) node;
	int attribute=predicateNode->getAttribute();
	TigrString &name=namesFile.getName(attribute);
	os << name;
	switch((*schema)[attribute]->getType())
	  {
	  case DISCRETE:
	    os << "=" << predicateNode->getIntValue() << ":"
	       << endl;
	    break;
	  case CONTINUOUS:
	    os << "<=" << predicateNode->getFloatValue()
	       << ":" << endl;
	    break;
	  }
	printSubtree(predicateNode->getLeftChild(),
		     numTabs+1,os);
	printSubtree(predicateNode->getRightChild(),
		     numTabs+1,os);
      }
      break;
    case TEMP_NODE:
      throw TigrStacktrace("This shouldn't happen");
    }
}



void Application::runTestCases(TrainingSet *testSet)
{  
  int TP=0, TN=0, FP=0, FN=0;
  int numCorrect=0, totalError=0;
  int i, n=testSet->numCases();
  for(i=0 ; i<n ; ++i)
    {
      TrainingCase &c=testSet->getCase(i);
      int actual=c.getCategory();
      int predicted=classifier->classify(c);
      
      if(timeSeries)
	{
	  int error=abs(actual-predicted);
	  totalError+=error;
	}
      else if(actual==predicted)
	++numCorrect;
      switch(predicted)
	{
	case 0:
	  switch(actual)
	    {
	    case 0: ++TN; break;
	    case 1: ++FN; break;
	    }
	  break;
	case 1:
	  switch(actual)
	    {
	    case 0: ++FP; break;
	    case 1: ++TP; break;
	    }
	  break;
	}
    }
  
  int numCategories=schema->getNumCategories();
  if(timeSeries)
    {
      int maxError=n*(numCategories-1);
      float accuracy=int((maxError-totalError)/double(maxError)*10000+0.5)/100;
      cout << "accuracy=" << accuracy << "%" << endl;
    }
  else
    {
      percentCorrect=100.0*numCorrect/(float)n;
      cout << "\t" << numCorrect 
	   << " correct out of " << n 
	   << " (" << percentCorrect << "%)" << endl;
      
      float precision=TP/float(TP+FP);
      float recall=TP/float(TP+FN);
      cout << "\nTP=" << TP 
	   << "\tTN=" << TN
	   << "\tFP=" << FP
	   << "\tFN=" << FN 
	   << "\nprecision=" << precision
	   << "\trecall=" << recall
	   << "\taccuracy="<<int(0.5+percentCorrect)<<"%"
	   << "\t#cases=" << n
	   << endl;
    }
}




TreeNode *Application::prune(TreeNode *node)
{
  if(node->getType()==LEAF_NODE) return node;

  PredicateNode *pred=(PredicateNode *) node;
  if(pred->getOriginalH()<pruneThreshold)//####
    {
      LeafNode *leaf=new LeafNode(pred->getMostPopularCategory());
      delete pred;
      return leaf;
    }
  else
    {
      pred->attachLeftChild(prune(pred->getLeftChild()));
      pred->attachRightChild(prune(pred->getRightChild()));
      return pred;
    }
}



void Application::getPruneThresholds(DecisionTree &tree)
{
  TigrVector<float> entropies;
  tree.getEntropies(entropies);
  sort(entropies.begin(),entropies.end());
  
  int i, n=entropies.size();
  for(i=0 ; i<n-1 ; ++i)
    {
      if(entropies[i]==entropies[i+1]) continue;
      float threshold=(entropies[i]+entropies[i+1])/2;
      thresholds.push_back(threshold);
    }
}




