/****************************************************************
 gp.C
 bmajoros@tigr.org

 Genetic programming.  See books by Koza for description of method.

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "Framework/Tableau.H"
#include "Framework/IndividualSchema.H"
#include "Framework/SymbolSet.H"
#include "Framework/CrosserOver.H"
#include "Framework/RouletteWheel.H"
#include "Framework/Mutator.H"
#include "Framework/SubtreeMutator.H"
#include "Framework/ClassifiedAR.H"
#include "Framework/PopulationBuilder.H"
#include "Framework/FitnessNormalizer.H"
#include "Framework/Stats.H"
#include "Framework/Tournament.H"
#include "../../NamesFile.H"
#include "../../TrainingSet.H"
#include "tigr++/TigrStrTokenizer.H"
#include "tigr++/TigrRandom.H"
#include "tigr++/TigrTime.H"
#include "tigr++/TigrCommandLine.H"
#include "tigr++/TigrConfigFile.H"
#include <iostream>
#include <stdlib.h>
#include <fstream>
using namespace std;

enum SelectionType
{
  FITNESS_PROPORTIONATE,
  TOURNAMENT_SELECTION
};

class Application
{
  bool timeSeries;
  TigrConfigFile *configFile;
  TigrString filestem, configFilename;
  SymbolSet symbolSet;
  IndividualBuilder *individualBuilder;
  Population *population;
  CrosserOver crosserOver;
  Mutator *pointMutator;
  SubtreeMutator *subtreeMutator;
  NamesFile namesFile;
  TigrVector<ClassifiedAR*> trainingSet, testSet;
  int numCategories;
  FitnessNormalizer fitnessNormalizer;
  SelectionType selectionType;
  int tournamentSize;
  int initialHeight;
  Individual *bestOfRun;
  int TP, TN, FP, FN;

  void processConfig();
  void setup();
  void setupSchema();
  void setupSymbols();
  void loadTrainingSets();
  void convertToARs(TrainingSet *,TigrVector<ClassifiedAR*> &);
  void bigLoop();
  void evaluateFitness(TigrVector<ClassifiedAR*> &,float proportion);
  void exercise(Individual &,TigrVector<ClassifiedAR*> &ARs,float proportion);
  Population *generateNextPop();
  inline Individual *selectIndividual(RouletteWheel &);
  void finish();
public:
  int go(int argc,char *argv[]);
};


int main(int argc,char *argv[])
{
  try
    {
      Application app;
      return app.go(argc,argv);
    }
  catch(const TigrString &msg)
    {
      cerr << msg.c_str() << endl;
      return -1;
    }
  catch(const char *msg)
    {
      cerr << msg << endl;
      return -1;
    }
  catch(...)
    {
      cerr<<"unknown exception caught in main()"<<endl;
      return -1;
    }
}



int Application::go(int argc,char *argv[])
{
  // Process the command line
  TigrCommandLine cmd(argc,argv,"tc:");
  if(cmd.numArgs()!=1)
    throw "\ngp [-t] [-c <config-file>] <filestem>\n\n"
      "t = time series\n";
  filestem=cmd.arg(0);
  timeSeries=cmd.option('t');
  if(cmd.option('c'))
    configFilename=cmd.optParm('c');
  else
    configFilename=filestem+".gp";

  // Create initial population, etc.
  setup();

  // Loop through the generations
  bigLoop();

  // Report results
  finish();
}



void Application::setup()
{
  // Process the configuration file
  processConfig();
  initialHeight=Tableau::getGlobal().getInitialHeight();

  // Create schema describing individuals
  setupSchema();

  // Create terminal & nonterminal sets
  setupSymbols();
  individualBuilder=new IndividualBuilder(symbolSet);

  // Load training & test sets
  loadTrainingSets();

  // Create an initial population
  PopulationBuilder populationBuilder(symbolSet);
  const int M=Tableau::getGlobal().getM();
  population=populationBuilder.buildPopulation(M,2,initialHeight);

  // Other misc. initialization
  pointMutator=new Mutator(symbolSet);
  subtreeMutator=new SubtreeMutator(individualBuilder->getTreeBuilder());
}



void Application::loadTrainingSets()
{
  TigrString namesFilename=filestem+".names";
  Schema *schema=namesFile.processFile(namesFilename);
  numCategories=schema->getNumCategories();

  TigrString trainFilename=filestem+".data";
  TrainingSet *train=new TrainingSet(*schema,trainFilename);
  convertToARs(train,trainingSet);
  delete train;

  TigrString testFilename=filestem+".test";
  TrainingSet *test=new TrainingSet(*schema,testFilename);
  convertToARs(test,testSet);
  delete test;

  delete schema;
}



void Application::convertToARs(TrainingSet *ts,TigrVector<ClassifiedAR*> &ARs)
{
  int i, n=ts->numCases();
  for(i=0 ; i<n ; ++i)
    {
      TrainingCase &tc=ts->getCase(i);
      ARs.push_back(new ClassifiedAR(tc));
    }
}



void Application::setupSchema()
{
  // Create schema describing individuals

  Tableau &tableau=Tableau::getGlobal();
  const int maxCall=tableau.getMaxCall();
  IndividualSchema::allocateGlobal(maxCall);
  TigrString adfArities=tableau.getAdfArities();
  TigrString entryArities=tableau.getEntryArities();
  
  // First, the ADF's
  TigrStrTokenizer adfTokenizer(adfArities,",;");
  while(adfTokenizer.hasMoreTokens())
    {
      const char *token=adfTokenizer.nextToken();
		if(!isdigit(token[0])) continue;
      int arity=atoi(token);
      IndividualSchema::getGlobal().addADF(arity);
    }

  // Now the entry point(s)
  TigrStrTokenizer entryTokenizer(entryArities," \t,;-");
  while(entryTokenizer.hasMoreTokens())
    {
      const char *token=entryTokenizer.nextToken();
      int arity=atoi(token);
      IndividualSchema::getGlobal().addEntryPoint(arity);
    }
}


void Application::bigLoop()
{
  bestOfRun=NULL;

  TigrTime stopwatch;
  stopwatch.startCounting();
  const int G=Tableau::getGlobal().getG();
  int generation, almostG=int(0.75*G);
  float trainPortion=Tableau::getGlobal().getTrainPortion();
  for(generation=0 ; generation<G ; ++generation)
    {
      cerr << "processing generation " << generation << ": "; 
      cerr.flush();// ###

      // Evaluate fitness of population members
      if(generation==almostG) trainPortion=1.1;
      evaluateFitness(trainingSet,trainPortion);
      Stats stats(*population);
      cerr << stats << endl;

      // See if a new bestOfRun has been produced
      Individual *winner=population->getMostFit();
      if(!bestOfRun ||
	 winner->getAdjustedFitness() > bestOfRun->getAdjustedFitness())
	bestOfRun=winner->clone();

      // Generate the next population
      Population *newPopulation=generateNextPop();

      // Replace the old with the new
      delete population;
      population=newPopulation;
    }
  stopwatch.stopCounting();
  cout << stopwatch.elapsedTime() << endl;
}



Population *Application::generateNextPop()
{
  // Get relevant parameters
  Tableau &tableau=Tableau::getGlobal();
  const float Pc=tableau.getPc();
  const float Pm=tableau.getPm();
  const float subtreeMutation=tableau.getSubtreeMutation();
  const float immigration=tableau.getImmigration();
  const float cloning=tableau.getCloning();
  const int M=tableau.getM();

  // Allocate a new population
  Population *newPopulation=new Population();
  
  // Decide how many crossovers & mutations to perform
  const int numCrossoverPairs=int(Pc*M/2);
  const int pointMutations=int(Pm*M);
  const int subtreeMutations=int(subtreeMutation*M);
  const int numImmigrants=int(immigration*M);
  const int numCopies=int(cloning*M);
  if(2*numCrossoverPairs+pointMutations+subtreeMutations+numImmigrants+
     numCopies != M)
    throw "crossover+mutation+immigration+cloning must equal 1.0";

  // Create a roulette wheel which reflects relative fitnesses
  RouletteWheel rouletteWheel(*population);
  
  // Crossover
  int i;
  for(i=0 ; i<numCrossoverPairs ; ++i)
    {
      Individual *mother=selectIndividual(rouletteWheel);
      Individual *father=selectIndividual(rouletteWheel);
      while(mother==father) 
	father=selectIndividual(rouletteWheel);
      Individual *daughter, *son;
      crosserOver.crossover(mother,father,daughter,son);
      newPopulation->addMember(daughter);
      newPopulation->addMember(son);
    }

  // Point mutation
  for(i=0 ; i<pointMutations ; ++i)
    {
      Individual *lucky=selectIndividual(rouletteWheel);
      if(Tableau::getGlobal().getMaxHeight()>0)
	lucky->getTreeHeight(); // height is cached for later
      Individual *mutant=pointMutator->mutate(lucky);
      newPopulation->addMember(mutant);
    }

  // Subtree mutation
  for(i=0 ; i<subtreeMutations ; ++i)
    {
      Individual *lucky=selectIndividual(rouletteWheel);
      if(Tableau::getGlobal().getMaxHeight()>0)
	lucky->getTreeHeight(); // height is cached for later
      Individual *mutant=subtreeMutator->mutate(lucky);
      newPopulation->addMember(mutant);
    }

  // Immigration
  for(i=0 ; i<numImmigrants ; ++i)
    newPopulation->addMember(
      individualBuilder->buildIndividual(initialHeight));

  // Exploitation (copy)
  for(i=0 ; i<numCopies ; ++i)
    {
      Individual *lucky=selectIndividual(rouletteWheel);
      if(Tableau::getGlobal().getMaxHeight()>0)
	lucky->getTreeHeight(); // height is cached for later
      newPopulation->addMember(lucky->clone());
    }

  return newPopulation;
}



Individual *Application::selectIndividual(RouletteWheel &rouletteWheel)
{
  switch(selectionType)
    {
    case FITNESS_PROPORTIONATE:
      return population->getMember(rouletteWheel.spin());
    case TOURNAMENT_SELECTION:
      {
	Tournament tournament(*population,tournamentSize);
	return tournament.pickWinner();
      }
    }
}



void Application::processConfig()
{  
  //  string configFilename=filestem+".gp";
  configFile=new TigrConfigFile(configFilename);

  Tableau::processConfig(*configFile);

  int seed=Tableau::getGlobal().getSeed();
  if(seed) SeedRandomizer(unsigned(seed));
  else randomize();

  selectionType=
    Tableau::getGlobal().useTournament() ? TOURNAMENT_SELECTION :
    FITNESS_PROPORTIONATE;
  tournamentSize=Tableau::getGlobal().getTournamentSize();
  if(selectionType==TOURNAMENT_SELECTION && tournamentSize<=0)
    throw string("set tournament-size to a positive integer");
}



void Application::setupSymbols()
{
  // Nonterminals
  TigrString symbols=Tableau::getGlobal().getNonterminals();
  TigrStrTokenizer tokenizer(symbols,",");
  while(tokenizer.hasMoreTokens())
    {
      const char *token=tokenizer.nextToken();
      if(!strcmp(token,"+")) 
	symbolSet.addNonterminal(new PlusNode());
      else if(!strcmp(token,"-")) 
	symbolSet.addNonterminal(new MinusNode());
      else if(!strcmp(token,"*")) 
	symbolSet.addNonterminal(new TimesNode());
      else if(!strcmp(token,"/")) 
	symbolSet.addNonterminal(new DivideNode());
      else if(!strcmp(token,"sin")) 
	symbolSet.addNonterminal(new SinNode());
      else if(!strcmp(token,"cos")) 
	symbolSet.addNonterminal(new CosNode());
      else if(!strcmp(token,"exp")) 
	symbolSet.addNonterminal(new ExpNode());
      else if(!strcmp(token,"log")) 
	symbolSet.addNonterminal(new LogNode());
      else if(!strcmp(token,"if")) 
	symbolSet.addNonterminal(new IfNode());
      else if(!strcmp(token,"call")) 
	symbolSet.addNonterminal(new CallNode());
      else if(!strcmp(token,"min")) 
	symbolSet.addNonterminal(new MinNode());
      else if(!strcmp(token,"max")) 
	symbolSet.addNonterminal(new MaxNode());
      else if(!strcmp(token,"fmod")) 
	symbolSet.addNonterminal(new FmodNode());
      else if(!strcmp(token,"abs")) 
	symbolSet.addNonterminal(new AbsNode());
      else if(!strcmp(token,"ceil")) 
	symbolSet.addNonterminal(new CeilNode());
      else if(!strcmp(token,"hypot")) 
	symbolSet.addNonterminal(new HypotNode());
      else if(!strcmp(token,"tanh")) 
	symbolSet.addNonterminal(new TanhNode());
      else if(!strcmp(token,"sinh")) 
	symbolSet.addNonterminal(new SinhNode());
      else if(!strcmp(token,"cosh")) 
	symbolSet.addNonterminal(new CoshNode());
      else if(!strcmp(token,"atanh")) 
	symbolSet.addNonterminal(new AtanhNode());
      else if(!strcmp(token,"asinh")) 
	symbolSet.addNonterminal(new AsinhNode());
      else if(!strcmp(token,"acosh")) 
	symbolSet.addNonterminal(new AcoshNode());
      else if(!strcmp(token,"log2")) 
	symbolSet.addNonterminal(new Log2Node());
      else if(!strcmp(token,"log10")) 
	symbolSet.addNonterminal(new Log10Node());
      else if(!strcmp(token,"cbrt")) 
	symbolSet.addNonterminal(new CbrtNode());
      else if(!strcmp(token,"sqrt")) 
	symbolSet.addNonterminal(new CallNode());
      else if(!strcmp(token,"atan")) 
	symbolSet.addNonterminal(new AtanNode());
      else if(!strcmp(token,"asin")) 
	symbolSet.addNonterminal(new AsinNode());
      else if(!strcmp(token,"acos")) 
	symbolSet.addNonterminal(new AcosNode());
      else if(!strcmp(token,"tan")) 
	symbolSet.addNonterminal(new TanNode());
      else if(!strcmp(token,"^")) 
	symbolSet.addNonterminal(new PowNode());
      else if(!strcmp(token,"%")) 
	symbolSet.addNonterminal(new ModNode());
      else if(!strcmp(token,"&&")) 
	symbolSet.addNonterminal(new AndNode());
      else if(!strcmp(token,"||")) 
	symbolSet.addNonterminal(new OrNode());
      else if(!strcmp(token,"!=")) 
	symbolSet.addNonterminal(new NotEqualsNode());
      else if(!strcmp(token,">=")) 
	symbolSet.addNonterminal(new GreaterEqualsNode());
      else if(!strcmp(token,"<=")) 
	symbolSet.addNonterminal(new LessEqualsNode());
      else if(!strcmp(token,"!")) 
	symbolSet.addNonterminal(new NotNode());
      else if(!strcmp(token,"<")) 
	symbolSet.addNonterminal(new LessNode());
      else if(!strcmp(token,">")) 
	symbolSet.addNonterminal(new GreaterNode());
      else if(!strcmp(token,">0")) 
	symbolSet.addNonterminal(new GreaterThanZeroNode());
      else if(!strcmp(token,">1")) 
	symbolSet.addNonterminal(new GreaterThanOneNode());
      else if(!strcmp(token,"=0")) 
	symbolSet.addNonterminal(new EqualsZeroNode());
      else
	throw string("nonterminal symbol \"")+token+"\" is not recognized";
    }
  
  // Terminals
  const float minConst=Tableau::getGlobal().getMinConst();
  const float maxConst=Tableau::getGlobal().getMaxConst();
  symbols=Tableau::getGlobal().getTerminals();
  tokenizer.tokenizeThis(symbols.c_str());
  while(tokenizer.hasMoreTokens())
    {
      const char *token=tokenizer.nextToken();
      if(!strcmp(token,"const")) 
	symbolSet.addTerminal(new ConstantNode(minConst,maxConst));
      else if(!strcmp(token,"var")) 
	symbolSet.addTerminal(new VariableNode());
      else
	throw string("terminal symbol \"")+token+"\" is not recognized";
    }
}



void Application::finish()
{
  // Get best individual from entire run
  evaluateFitness(trainingSet,1.1);
  Individual *winner=population->removeMostFit();
  if(!bestOfRun ||
     winner->getAdjustedFitness() > bestOfRun->getAdjustedFitness())
    {
      delete bestOfRun;
      bestOfRun=winner;
    }

  // Test the winner on the test set
  delete population;
  population=new Population();
  population->addMember(bestOfRun);
  evaluateFitness(testSet,1.1);

  // Report results
  int testSetSize=testSet.size();
  float fitness=bestOfRun->getStandardizedFitness();
  float accuracy=int(1000*(1-fitness)+0.5)/10.0;
  cout << "accuracy of winner on test set: " << accuracy << "%" << endl;
  ofstream os((filestem+".gp.tree").c_str());
  os << *bestOfRun << endl;
  float pDenom=TP+FP, rDenom=TP+FN;
  float precision=pDenom ? TP/pDenom : 0;
  float recall=rDenom ? TP/rDenom : 0;
  cout << "TP=" << TP << "\nTN=" << TN << "\nFP=" << FP
		 << "\nFN=" << FN << "\nprecision=" << precision
		 << "\nrecall=" << recall << endl;
}



void Application::evaluateFitness(TigrVector<ClassifiedAR*> &ARs,
				  float proportion)
{
  int i, n=population->size();
  for(i=0 ; i<n ; ++i)
    {
      Individual *ind=population->getMember(i);
      exercise(*ind,ARs,proportion);
    }

  fitnessNormalizer.normalize(*population);
}



void Application::exercise(Individual &ind,TigrVector<ClassifiedAR*> &ARs,
			   float proportion)
{
  int errors=0;
  TP=TN=FP=FN=0;

  // Get the result-producing branch, so we can call it
  int resultBranch=Tableau::getGlobal().getResultBranch();
  TreeNode *tree=ind.getEntryPoint(resultBranch);

  // Iterate through the test cases
  int i, n=ARs.size(), num=0;
  for(i=0 ; i<n ; ++i)
    {
      // Exercise only on a random sample of cases
      if(Random0to1()>=proportion) continue;

      // Get the next test case
      ClassifiedAR *ar=ARs[i];

      // Call the evolved procedure
      float value=tree->eval(ar->getAR(),ind);

      // Discretize the result into a category
      int category=int(value+0.5);
      if(category<0) category=0;
      else if(category>numCategories-1) 
		  category=numCategories-1;

      // Score the result
      int actualCategory=ar->getCategory();
      if(category!=actualCategory)
	if(timeSeries)
	  errors+=abs(category-actualCategory);
	else
	  {
	    ++errors;
	    if(category==1) ++FP;
	    else ++FN;
	  }
      else
	if(category==1) ++TP;
	else ++TN;
      
      ++num;
    }

  double standardized;
  if(timeSeries)
    {
      int maxErrors=num*(numCategories-1);
      standardized=errors/double(maxErrors);
    }
  else
    standardized=errors/float(num);
  ind.setStandardizedFitness(standardized);
  ind.setAdjustedFitness(1.0/(1.0+errors));
}


