/****************************************************************
 knn.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include <iostream>
#include <stdlib.h>
#include <fstream>
#include <math.h>
#include <unistd.h>
#include "tigr++/TigrExceptions.H"
#include "tigr++/TigrString.H"
#include "tigr++/TigrStack.H"
#include "tigr++/TigrStacktrace.H"
#include "tigr++/TigrBitSet.H"
#include "tigr++/TigrVector.H"
#include "tigr++/TigrVectorSorter.H"
#include "tigr++/TigrRandom.H"
#include "tigr++/TigrArray1D.H"
#include "tigr++/TigrArray2D.H"
#include "tigr++/TigrCommandLine.H"
#include "tigr++/TigrInversionMatrix.H"
#include "tigr++/TigrGaussJordan.H"
#include "tigr++/TigrTime.H"
#include "tigr++/BoundedSortedArray.H"
#include "tigr++/Constants.H"
#include "../NamesFile.H"
#include "../TrainingSet.H"
#include "../Schema.H"
#include "FeatureF.H"
using namespace std;

class Application
{
  int numCategories;
  TigrString filestem;
  NamesFile namesFile;
  TrainingSet *trainingSet, *testSet;
  TigrBitSet *features; // Which attributes are included in model
  TigrBitSet *trainingTraining;// For cross-validation
  TigrVector<TigrVector<TrainingCase*>*> groups; // each group is a vector of TrainingCase
  TigrVector<FeatureF*> f_values; // vector of FeatureF
  bool stepwise, weighted, useMahalanobis, timeSeries;
  TigrDblArray1D weights;
  TigrDblMatrix *inverseCovMatrix;

  Schema *schema;
  double categoryWeight;
  int K;

  double classify();
  double distance(TrainingCase &,TrainingCase &);
  void getNeighbors(TrainingCase &,BoundedSortedArray<TrainingCase*,double> &);
  int predictCategory(BoundedSortedArray<TrainingCase*,double> &);
  double computeFRatio(int attrNum);
  void setUpGroups();
  int groupSize(int groupNum);
  void featureSelection();
  void computeFratios();
  double crossVal_classify();
  void crossVal_getNeighbors(TrainingCase &,
			     BoundedSortedArray<TrainingCase*,double> &);
  void computeCovMatrix();
  double mahalanobis(TrainingCase &,TrainingCase &);
public:
  Application(TigrString filestem,int K,bool stepwise,
	      bool weighted,bool useMahalanobis,bool timeSeries);
  void go();
};


void usage();


int main(int argc,char *argv[])
{
  try
    {
      TigrTime stopwatch;
      stopwatch.startCounting();
      randomize();

      TigrCommandLine cmd(argc,argv,"wsmt");
      if(cmd.numArgs()!=2) usage();
      bool stepwise=cmd.option('s');
      bool weighted=cmd.option('w');
      bool useMahalanobis=cmd.option('m');
      bool timeSeries=cmd.option('t');

      Application app(argv[optind+1],atoi(argv[optind]),
		      stepwise,weighted,useMahalanobis,
		      timeSeries);
      app.go();
      stopwatch.stopCounting();
      cout << stopwatch.elapsedTime() << endl;
    }
  catch(RootException &e)
    {
      cerr << e.getMessage() << endl;
      return -1;
    }
  catch(TigrString s)
    {
      cerr << s.c_str() << endl;
      return -1;
    }
  catch(const char *s)
    {
      cerr << s << endl;
      return -1;
    }

  return 0;
}



void usage()
{
  throw TigrString("\
knn [-wsm] <K> <filestem>\n\
   -w : weight variables by F ratio (between-group MS/within-group MS)\n\
   -s : stepwise - drop variables with low F ratio\n\
   -m : mahalanobis - account for multicollinearity\n\
   -t : time series - score accuracy using category distance\n\
\n");
}



Application::Application(TigrString filestem,int K,bool stepwise,
			 bool weighted,bool useMahalanobis,
			 bool timeSeries)
  : filestem(filestem), 
    trainingSet(NULL), 
    weighted(weighted),
    categoryWeight(1.0), 
    K(K), 
    weights(0),
    stepwise(stepwise), 
    useMahalanobis(useMahalanobis), 
    timeSeries(timeSeries)
{
  // ctor
}



void Application::go()
{
  TigrString namesFilename=filestem+".names";
  TigrString trainFilename=filestem+".data";
  TigrString testFilename=filestem+".test";

  schema=namesFile.processFile(namesFilename);
  numCategories=schema->getNumCategories();
  features=new TigrBitSet(schema->numAttributes());
  features->addAll();
  weights.resize(schema->numAttributes());

  trainingSet=new TrainingSet(*schema,trainFilename);
  int numTrainingCases=trainingSet->numCases();
  trainingTraining=new TigrBitSet(numTrainingCases);
  setUpGroups();
  computeFratios();
  if(useMahalanobis) computeCovMatrix();
  if(stepwise) featureSelection();

  cerr << "Loading test set..." << endl;
  testSet=new TrainingSet(*schema,testFilename);

  cerr << "Classifying test cases..." << endl;
  double accuracy=classify();
  cout << "\n\naccuracy on test set: " << accuracy << endl;
}



double Application::classify()
{
  int totalError=0;
  double lastPercent=0.0;
  int n=testSet->numCases();
  int TP=0, TN=0, FP=0, FN=0;
  for(int i=0 ; i<n ; ++i)
    {
      double percentDone=100.0 * i / n;
      if(percentDone>=lastPercent+5)
	{
	  cout << int(percentDone) << "% "; cout.flush();
	  lastPercent=percentDone;
	}
      TrainingCase &thisCase=testSet->getCase(i);
      
      BoundedSortedArray<TrainingCase*,double> neighbors(K,KEEP_LOW_VALUES);
      getNeighbors(thisCase,neighbors);
      int predicted=predictCategory(neighbors);
      int expected=thisCase.getCategory();
      
      if(timeSeries)
	totalError+=abs(expected-predicted);
      else
	{
	  switch(expected)
	    {
	    case 0:
	      switch(predicted)
		{
		case 0: ++TN;	break;
		case 1:	++FP;	break;
		}
	      break;
	    case 1: 
	      switch(predicted)
		{
		case 0:	++FN;	break;
		case 1:	++TP;	break;
		}
	      break;
	    }
	}
    }
  int numCorrect=TP+TN;
  
  float precision=TP/float(TP+FP);
  float recall=TP/float(TP+FN);
  cout << "\nTP=" << TP 
       << "\nTN=" << TN
       << "\nFP=" << FP
       << "\nFN=" << FN 
       << "\nprecision=" << precision
       << "\nrecall=" << recall
       << "\n#cases=" << n
       << endl;
  
  double accuracy;
  if(timeSeries)
    {
      int maxError=n*(numCategories-1);
      accuracy=100*(maxError-totalError)/double(maxError);
    }
  else
    accuracy=100.0 * numCorrect/(double) n;
  return accuracy;
}



void Application::getNeighbors(TrainingCase &thisCase,
		BoundedSortedArray<TrainingCase*,double> &neighbors)
{
  int i;
  for(i=0 ; i<trainingSet->numCases() ; ++i)
    {
      TrainingCase &thatCase=trainingSet->getCase(i);
      double d=distance(thisCase,thatCase);
      neighbors.add(&thatCase,d);
    }
}



int Application::predictCategory(
        BoundedSortedArray<TrainingCase*,double> &neighbors)
{
  TigrDblArray1D distances(numCategories);
  distances.setAllTo(0.0);
  TigrBitSet touched(numCategories);

  for(int i=0 ; i<K ; ++i)
    {
      TrainingCase &c=*neighbors[i];
      int category=c.getCategory();
      if(category>=numCategories) 
	throw TigrString("Illegal category: ")+category;
      double distance=neighbors.getValue(i);
      distances[category]+=1/(distance*distance);
      touched.addMember(category);
    }

  int lowestIndex=-1;
  double lowestDist=0;//FLT_MAX;
  for(int i=0 ; i<numCategories ; ++i)
    if(touched.isMember(i))
       {
	 double d=distances[i];
	 if(d >/*### < */ lowestDist)
	   {
	     lowestIndex=i;
	     lowestDist=d;
	   }
       }

  return lowestIndex;
}



double Application::distance(TrainingCase &case1,TrainingCase &case2)
{
  // Computes euclidean distance

  if(useMahalanobis) return mahalanobis(case1,case2);

  double sum=0;

  int i, n=schema->numAttributes();
  for(i=0 ; i<n ; ++i)
    {
      // Make sure this feature hasn't been dropped
      // (due to a low F value)
      if(!features->isMember(i)) continue;

      switch((*schema)[i]->getType())
	{
	case DISCRETE:
	  if(case1.getDiscrete(i)!=case2.getDiscrete(i))
	    sum+=weights[i];//categoryWeight;
	  break;
	case CONTINUOUS:
	  {
	    double delta=
	      case1.getContinuous(i)-case2.getContinuous(i);
	    if(weighted)
	      sum+=delta*delta*weights[i];
	    else
	      sum+=delta*delta;
	  }
	  break;
	}
    }

  return sqrt(sum);
}



double Application::computeFRatio(int attrNum)
{
  /*
    Computes MS(groups)/MS(error) = between-group 
    variability vs. within-group variability = a measure of
    how well this variable separates groups.
   */

  double sumX=0.0;
  int k=schema->getNumCategories();
  int i, N=0;
  double totalSS=0.0, groupSS=0.0;

  for(i=0 ; i<k ; ++i)
    {
      TigrVector<TrainingCase*> *group=groups[i];
      int groupSize=group->size();
      if(groupSize==0) continue;
      N+=groupSize;

      double localSumX=0.0;
      int j;
      for(j=0 ; j<groupSize ; ++j)
	{
	  TrainingCase *thisCase=(*group)[j];
	  double x=thisCase->getContinuous(attrNum);
	  localSumX+=x;
	  totalSS+=(x*x);
	}
      sumX+=localSumX;
      groupSS+=(localSumX/groupSize*localSumX);
    }

  double C=sumX/N*sumX;
  totalSS-=C;
  groupSS-=C;
  double errorSS=totalSS-groupSS;
  double totalDF=N-1;
  double groupDF=k-1;
  double errorDF=totalDF-groupDF;
  double groupMS=groupSS/groupDF;
  double errorMS=errorSS/errorDF;
  
  double ratio;
  if(errorMS==0)
    {
      ratio=POSITIVE_INFINITY;
      cerr << "WARNING!  F ratio=INF (" << ratio << ")" << endl; // ###
    }
  else
    ratio=groupMS/errorMS;
  return ratio;
}



void Application::setUpGroups()
{
  int nGroups=schema->getNumCategories();
  int i;
  for(i=0 ; i<nGroups ; ++i)
    groups.push_back(new TigrVector<TrainingCase*>);

  int n=trainingSet->numCases();
  for(i=0 ; i<n ; ++i)
    {
      TrainingCase &c=trainingSet->getCase(i);
      int cat=c.getCategory();
      if(cat>=nGroups) cat=nGroups-1;
      TigrVector<TrainingCase*> *group=groups[cat];
      group->push_back(&c);
    }
}



int Application::groupSize(int groupNum)
{
  return groups[groupNum]->size();
}



void Application::computeFratios()
{
  int numFeatures=schema->numAttributes();
  int i;
  for(i=0 ; i<numFeatures ; ++i)
    {
      double F=computeFRatio(i);
      //cerr << "F(" << i << ")=" << F << endl;
      
      f_values.push_back(new FeatureF(i,F));
      weights[i]=1/F;
    }
}



void Application::featureSelection()
{
  cerr << "Performing feature selection..." << endl; 

  int numFeatures=schema->numAttributes();

  // Sort the F values in increasing order
  FeatureComparator cmp;
  TigrVectorSorter<FeatureF*> sorter(f_values,cmp);
  sorter.sortAscendInPlace();

  // Split training set for cross-validation
  trainingTraining->purge();
  int i, trainingSetSize=trainingSet->numCases();
  for(i=0 ; i<trainingSetSize ; ++i)
    if(Random0to1()<0.66)
      trainingTraining->addMember(i);

  // Compute cross-val accuracy with all features left
  // in the model
  double bestAccuracy=crossVal_classify();
  cerr << "acc=" << bestAccuracy << endl; // ###

  // Try removing some features to see if accuracy is
  // improved
  for(i=0 ; i<numFeatures ; ++i)
    {
      FeatureF *ff=f_values[i];
      features->removeMember(ff->featureNum);
      
      double accuracy=crossVal_classify();
      if(accuracy<bestAccuracy)
	{
	  cerr << "Removing feature " << ff->featureNum
	       << " reduces acc to " << accuracy
	       << ", so it will not be removed." << endl; // ###
	  features->addMember(ff->featureNum);
	  break;
	}
      bestAccuracy=accuracy;
      cerr << "removing feature " << ff->featureNum
	   << ".  Now acc=" << bestAccuracy << endl; // ###
    }
}



double Application::crossVal_classify()
{
  /*
    This is the cross-validation version of classify()
   */

  int numCorrect=0, totalError=0;

  int n=trainingSet->numCases();
  for(int i=0 ; i<n ; ++i)
    if(!trainingTraining->isMember(i))
      {
	TrainingCase &thisCase=trainingSet->getCase(i);
	
	BoundedSortedArray<TrainingCase*,double> neighbors(K,KEEP_LOW_VALUES);
	crossVal_getNeighbors(thisCase,neighbors);
	int predictedCategory=predictCategory(neighbors);
	
	if(timeSeries)
	  totalError+=abs(thisCase.getCategory()-predictedCategory);
	else
	  if(thisCase.getCategory()==predictedCategory)
	    ++numCorrect;
      }

  double accuracy;
  if(timeSeries)
    {
      int maxError=n*(numCategories-1);
      accuracy=100*(maxError-totalError)/double(maxError);
    }
  else
    accuracy=100.0 * numCorrect/(double) n;

  return accuracy;
}



void Application::crossVal_getNeighbors(TrainingCase &thisCase,
	       BoundedSortedArray<TrainingCase*,double> &neighbors)
{
  /*
    This is the cross-validation version of getNeighbors()
   */

  for(int i=0 ; i<trainingSet->numCases() ; ++i)
    if(trainingTraining->isMember(i))
      {
	TrainingCase &thatCase=trainingSet->getCase(i);
	double d=distance(thisCase,thatCase);
	neighbors.add(&thatCase,d);
      }
}



void Application::computeCovMatrix()
{
  int numAttr=schema->numAttributes();
  TigrDblArray2D sumXY(numAttr,numAttr);
  TigrDblArray1D sumX(numAttr);

  sumXY.setAllTo(0.0);
  sumX.setAllTo(0.0);

  int numCases=trainingSet->numCases();
  for(int i=0 ; i<numCases ; ++i)
    {
      TrainingCase &thisCase=trainingSet->getCase(i);
      for(int j=0 ; j<numAttr ; ++j)
	{
	  double x=thisCase.getContinuous(j);
	  sumX[j]+=x;
	  for(int k=0 ; k<numAttr ; ++k)
	    {
	      double y=thisCase.getContinuous(k);
	      sumXY[j][k]+=x*y;
	    }
	}
    }
  
  TigrDblMatrix covMatrix(numAttr,numAttr);
  for(int j=0 ; j<numAttr ; ++j)
    for(int k=0 ; k<numAttr ; ++k)
      {
	covMatrix(j,k)=
	  (sumXY[j][k]-sumX[j]*sumX[k]/numCases)/
	  (numCases-1.0);
      }

  inverseCovMatrix=new TigrDblMatrix(numAttr,numAttr);
  if(!TigrGaussJordan::invert(covMatrix,*inverseCovMatrix))
    {
      cerr << "WARNING: cov matrix not invertible"
	   << "\n         using euclidean distance" << endl;
      useMahalanobis=false;
    }
}



double Application::mahalanobis(TrainingCase &case1,
			       TrainingCase &case2)
{
  int numAttr=schema->numAttributes();
  TigrDblMatrix left(1,numAttr), right(numAttr,1);

  int i;
  for(i=0 ; i<numAttr ; ++i)
    {
      double d=case1.getContinuous(i)-case2.getContinuous(i);
      if(weighted)
	left(0,i)=right(i,0)=d*sqrt(weights[i]);
      else
	left(0,i)=right(i,0)=d;
    }

  TigrDblMatrix temp1(1,numAttr);
  multiply(left,*inverseCovMatrix,temp1);

  TigrDblMatrix temp2(1,1);
  multiply(temp1,right,temp2);

  double d=temp2(0,0);
  return d;
}


