/****************************************************************
 bayes.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include <iostream>
#include "tigr++/TigrExceptions.H"
#include "tigr++/TigrCommandLine.H"
#include "../Schema.H"
#include "../TrainingSet.H"
#include "../NamesFile.H"
#include "ModelBuilder.H"
using namespace std;

int NUM_BUCKETS; // for discretization of continuous attributes


// ######################################################

class Application
{
public:
  void go(int argc,char *argv[]);
  void validate(Model &,TrainingSet &,const char *label,
		int &TP,int &TN,int &FP,int &FN);
};

// ######################################################

int main(int argc,char *argv[])
{
  try
    {
      Application app;
      app.go(argc,argv);
    }
  catch(RootException &e)
    {
      cerr << e.getMessage() << endl;
      return -1;
    }
  catch(const TigrString &s)
    {
      cerr << "Exception: "<<s<<endl;
      return -1;
    }
  catch(const char *p)
    {
      cerr<<"Exception: "<<p<<endl;
      return -1;
    }
  catch(...)
    {
      cerr << "Unknown exception caught in main()" << endl;
      return -1;
    }
  return 0;
}

// ######################################################

void Application::go(int argc,char *argv[])
{
  TigrCommandLine cmd(argc,argv,"");
  if(cmd.numArgs()!=2)
    throw "bayes <filestem> <#discretization-bins>";
  TigrString filestem=cmd.arg(0);
  NUM_BUCKETS=cmd.arg(1).asInt();
  TigrString configFilename=filestem+".names";
  TigrString trainingDataFilename=filestem+".data";
  TigrString testDataFilename=filestem+".test";

  // Load configuration
  //cerr << "Loading configuration..." << endl;
  NamesFile namesFile;
  Schema &schema=*namesFile.processFile(configFilename);

  // Load training data
  //cerr << "Loading training data..." << endl;
  TrainingSet trainingData(schema,trainingDataFilename);

  // Build model
  //cerr << "Building model..." << endl;
  ModelBuilder modelBuilder(trainingData,schema);
  Model *model=modelBuilder.buildModel();

  // Validate on training data
  //cerr << "Classifying training data..." << endl;
  int TP=0, TN=0, FP=0, FN=0;
  validate(*model,trainingData,"train: ",TP,TN,FP,FN);

  // Load test data
  //cerr << "Loading test data..." << endl;
  TrainingSet testData(schema,testDataFilename);

  // Classify test data
  //cerr << "Processing test data..." << endl;
  TP=TN=FP=FN=0;
  validate(*model,testData,"test: ",TP,TN,FP,FN);
  float precision=TP/float(TP+FP);
  float recall=TP/float(TP+FN);
  cout << "TP=" << TP 
       << "\nTN=" << TN
       << "\nFP=" << FP
       << "\nFN=" << FN 
       << "\nprecision=" << precision
       << "\nrecall=" << recall
       << "\n#cases=" << testData.numCases()
       << endl;
}



void Application::validate(Model &model,TrainingSet &dataSet,
			   const char *label,
			   int &TP,int &TN,int &FP,int &FN)
{
  int right=0;
  
  int n=dataSet.numCases();
  for(int i=0 ; i<n ; ++i)
    {
      TrainingCase &thisCase=dataSet.getCase(i);
      const int expected=thisCase.getCategory();
      const int computed=model(thisCase);
      if(computed==expected) ++right;
      switch(computed)
	{
	case 0:
	  switch(expected)
	    {
	    case 0: ++TN; break;
	    case 1: ++FN; break;
	    }
	  break;
	case 1:
	  switch(expected)
	    {
	    case 0: ++FP; break;
	    case 1: ++TP; break;
	    }
	  break;
	}
    }
  
  const float percentCorrect=100.0*float(right)/n;
  cout << label << percentCorrect << "% accuracy" << endl;
}


