/****************************************************************
 classify.C
 bmajoros@tigr.org

 A pre-packaged neural-network app that reads an arbitrary data 
 file, trains a neural network on that data, and then optionally 
 tests the network on a separate test file.  Employs multiple 
 output neurons for classification, one for each category.

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/

/*********************************************************
  classify.C : 
  5/19/1999 : created (B. Majoros)
  5/13/2000 : 
  12/17/2003 : Updated to use tigr++ class library
******************************************************/

// ---------------- includes ------------------
#include <iostream>
#include <fstream>
#include <math.h>
#include <stdlib.h>
#include "tigr++/TigrCommandLine.H"
#include "tigr++/TigrRandom.H"
#include "tigr++/TigrConfigFile.H"
#include "tigr++/TigrStacktrace.H"
#include "tigr++/TigrVector.H"
#include "tigr++/TigrLinRegressor.H"
#include "tigr++/TigrSummaryStats.H"
#include "tigr++/TigrVector.H"
#include "tigr++/TigrTime.H"
#include "tigr++/TigrExceptions.H"
#include "NeuralNetwork.H"
#include "StandardNeuron.H"
#include "NeuralCase.H"
#include "Backprop.H"
#include "NetworkBuilder.H"
#include "Normalizer.H"
#include "../NamesFile.H"
using namespace std;

// --------------- globals -----------------------
ofstream ofs("log.txt");

class Application
{
  TigrString testOutput, trainOutput, backpropLog;
  TigrString dataFilename, testFilename, configFilename;
  TigrVector<NeuralCase*> trainingCases;
  TigrVector<NeuralCase*> testCases;
  DoubleVector outputRange;
  float noiseFactor;
  NamesFile namesFile;
  Schema *schema;

  bool shouldRandomize;
  Normalizer *normalizer;
  NeuralNetwork *nn;
  int maxIterations;
  float learningRate;
  int numLayers;
  int neuronsPerLayer;
  int numInputNodes;
  float maxAdj, minAdj;
  int numCategories;

  void buildFilenames(const TigrString &);
  bool loadData(TigrString &filename,
		TigrVector<NeuralCase*> &);
  void buildNetwork();
  void trainNetwork();
  void testNetwork(TigrVector<NeuralCase*> &cases,
		   TigrString &filename);
  void buildNeuralCases(TigrVector<NeuralCase*> &);
  void testNetwork(TigrVector<NeuralCase*> &,
		   IntVector &outputIndices,ostream &);
  void applyCase(NeuralCase &c);
  void processConfigFile();
public:
  Application();
  void go(int argc,char *argv[]);
};



int main(int argc,char *argv[])
{
  try
    {
      TigrTime stopwatch;
      stopwatch.startCounting();

      Application app;
      app.go(argc,argv);

      stopwatch.stopCounting();
      cout << stopwatch.elapsedTime() << endl;
      return 0;
    }
  catch(RootException &e)
    {
      cerr << "Exception: " << e.getMessage() << endl;
      return -1;
    }
  catch(const char *msg)
    {
      cerr << "exception: " << msg << endl;
      return -1;
    }
  catch(char *msg)
    {
      cerr << "exception: " << msg << endl;
      return -1;
    }
  catch(string s)
    {
      cerr << s.c_str() << endl;
      return -1;
    }
  catch(...)
    {
      cerr << "unknown exception caught in main()" << endl;
      return -1;
    }
}



Application::Application()
  : normalizer(NULL) 
{
  // ctor
}



void Application::go(int argc,char *argv[])
{
  TigrCommandLine cmd(argc,argv,"");
  if(cmd.numArgs()!=1) throw "classify <filestem>";
  TigrString filestem=cmd.arg(0);

  buildFilenames(filestem);
  schema=namesFile.processFile(filestem+".names");
  numCategories=schema->getNumCategories();

  processConfigFile();

  if(!loadData(dataFilename,trainingCases)) 
    throw "can't load training data";
  loadData(testFilename,testCases);
  
  if(shouldRandomize) randomize();

  buildNetwork();

  trainNetwork();

  cout<<"training set: ";
  testNetwork(trainingCases,trainOutput);
  if(!testCases.isEmpty())
    {
      cout<<"test set: ";
      testNetwork(testCases,testOutput);
    }
}



void Application::buildFilenames(const TigrString &filestem)
{
  testOutput=filestem+".test.out";
  trainOutput=filestem+".train.out";
  backpropLog=filestem+".backprop.log";
  dataFilename=filestem+".data";
  testFilename=filestem+".test";
  configFilename=filestem+".config";
}



bool Application::loadData(TigrString &filename,
			   TigrVector<NeuralCase*> &cases)
{
  TigrString line;
  DoubleVector vector;

  ifstream is(filename.c_str());
  if(!is.good()) return false;
  while(!is.eof())
    {
      line.getline(is);
      if(line.isEmpty()) break;

      vector.clear();

      TigrStrTokenizer tokenizer(line,", \t");
      while(tokenizer.hasMoreTokens())
	{
	  const char *token=tokenizer.nextToken();
	  double value=atof(token);
	  vector.push_back(value);
	}
      if(vector.isEmpty()) break;

      int numInputs=vector.size()-1;

      NeuralCase *c=new NeuralCase(numInputs,numCategories);

      for(int i=0 ; i<numInputs ; ++i)
	{
	  double value=vector[i];
	  c->setInput(i,(float) value);
	}

      int category=int(vector[numInputs]);
      for(int i=0 ; i<numCategories ; ++i)
	c->setOutput(i,i==category ? 0.9 : 0.1);
      c->setCategory(category);

      cases.push_back(c);

      numInputNodes=numInputs;
    }

  return true;
}



void Application::processConfigFile()
{
  TigrConfigFile configFile(configFilename);
  maxIterations=configFile.getIntOrDie("maxIterations");
  cout << "maxIterations=" << maxIterations << endl;

  learningRate=configFile.getFloatOrDie("learningRate");
  cout << "learningRate=" << learningRate << endl;

  numLayers=configFile.getIntOrDie("numLayers");
  cout << "numLayers=" << numLayers << endl;

  neuronsPerLayer=configFile.getIntOrDie("neuronsPerLayer");
  cout << "neuronsPerLayer=" << neuronsPerLayer << endl;

  minAdj=configFile.getFloatOrDie("min-adj");
  cout << "minAdj is " << minAdj << endl;

  maxAdj=configFile.getFloatOrDie("max-adj");
  cout << "maxAdj" << maxAdj << endl;

  shouldRandomize=(configFile.getIntOrDie("randomize") ? true : false);
  cout << "shouldRandomize is"
       << (shouldRandomize ? "true" : "false") << endl;

  noiseFactor=configFile.getFloatOrDie("noise-factor");
  cout << "noiseFactor: " << noiseFactor << endl;
}




void Application::buildNetwork()
{
  nn=NetworkBuilder::buildNetwork(numInputNodes,
     numCategories,numLayers,neuronsPerLayer);
}



void Application::trainNetwork()
{
  Backpropagator trainer;

  cout << "normalizing..." << endl;
  normalizer=new Normalizer(numInputNodes,0);
  normalizer->normalize(trainingCases,maxAdj,minAdj);
  if(!testCases.isEmpty())
    normalizer->normalize(testCases,maxAdj,minAdj);
  cout << "done." << endl;

  cout << "training..." << endl;
  LogisticFunction transferFunction;
  ofstream backpropLog(backpropLog.c_str());

  trainer.train(*nn,trainingCases,learningRate,
		transferFunction,maxIterations,
		backpropLog,noiseFactor);
  backpropLog.close();
  cout << "done." << endl;
}




void Application::testNetwork(TigrVector<NeuralCase*> 
			      &cases,TigrString &filename)
{
  ofstream os(filename.c_str());

  IntVector &outputIndices=nn->getOutputIndices();
  
  testNetwork(cases,outputIndices,os);
}



void Application::testNetwork(TigrVector<NeuralCase*> &
			      cases,IntVector &outputIndices,
			      ostream &os)
{
  int numCorrect=0;

  int i;
  for(i=0 ; i<cases.size() ; ++i)
    {
      NeuralCase *c=(NeuralCase *) cases[i];
      applyCase(*c);

      // Get the actual category
      int expected=c->getCategory();
      os << "\ttru=" << expected;

      // Get the neural network's prediction
      int j;
      double largestValue=nn->getOutput(outputIndices[0]);
      int largestIndex=0;
      for(j=0 ; j<numCategories ; ++j)
	{
	  double value=nn->getOutput(outputIndices[j]);
	  if(value>largestValue)
	    {
	      largestValue=value;
	      largestIndex=j;
	    }
	}
      int computed=largestIndex;

      os << "\tnet=" << computed;

      if(computed==expected) ++numCorrect;

      os << endl;
    }
  os << "=========================================="
    << endl;

  float accuracy=int(100*numCorrect/(float) cases.size()+0.5);
  cout << "accuracy=" << accuracy << "%" << endl;

  // Display options
  os << "iterations:               " << maxIterations 
     << endl;
  os << "num hidden layers:        " << numLayers << endl;
  os << "neurons per hidden layer: " << neuronsPerLayer
     << endl;
  os << "learning rate:            " << learningRate 
     << endl;
}



void Application::applyCase(NeuralCase &c)
{
  /* 
     This function simply applies the inputs of the 
     training case to the input neurons and then 
     propagates the network to compute its outputs.
     */

  // Apply inputs to input neurons
  IntVector &inputIndices=nn->getInputIndices();
  int i;
  for(i=0 ; i<inputIndices.size() ; ++i)
    {
      int index=inputIndices[i];
      float value=c.getInput(i);

      nn->setInput(index,value);
    }

  // Propagate the system
  nn->propagate();
}

