/**************************************************************
TigrStringMap.H
bmajoros@tigr.org 1/1/2003

TIGR++ : C++ class template library for bioinformatics

Copyright (c) 2003, The Institute for Genomic Research (TIGR), 
Rockville, Maryland, U.S.A.  All rights reserved.
***************************************************************/
#ifndef INCL_TigrStringMap_H
#define INCL_TigrStringMap_H
#include <list>
#include "TigrString.H"
using namespace std;

/*****************************************************************
 */
template<class T>
struct StringMapElem
{
  int len;
  char *first;
  T second;
  StringMapElem(const char *,int len,const T &);
  StringMapElem(const char *,int len);
  StringMapElem(const StringMapElem<T> &);
  virtual ~StringMapElem();
};

/*****************************************************************
 */
template<class T>
class StringMapIterator
{
public:
  typedef list<StringMapElem<T>*> Bucket;
  typedef StringMapElem<T> ElemType;
  StringMapIterator(int index,int tableSize,Bucket *array);
  inline bool operator!=(const StringMapIterator<T> &);
  bool operator==(const StringMapIterator<T> &);
  StringMapIterator<T> &operator++(int);
  StringMapIterator<T> &operator++();
  StringMapElem<T> &operator*();
private:
  int index, tableSize;
  typename Bucket::iterator cur, end;
  Bucket *array;
  void findNonemptyBucket();
};

/*****************************************************************
 */
template<class T>
class TigrStringMap
{
public:
  typedef StringMapElem<T> ElemType;
  typedef StringMapIterator<T> iterator;
  typedef StringMapIterator<T> const_iterator;
  TigrStringMap(int tableSize);
  TigrStringMap(const TigrStringMap<T> &);
  virtual ~TigrStringMap();
  T &lookup(const char *,int index,int len);
  T &lookup(const char *,int len);
  bool isDefined(const char *,int index,int len);
  bool isDefined(const char *,int len);
  int size();
  iterator begin();
  iterator end();
  iterator begin() const;
  iterator end() const;
  void clear();
  void remove(const char *,int index,int len);
  void remove(const char *,int len);
  TigrStringMap<T> &operator=(const TigrStringMap &);
private:
  typedef list<StringMapElem<T>*> Bucket;
  int tableSize, numElements;
  Bucket *array;
  unsigned hash(const char *,int len,int tableSize);
};



/*****************************************************************
 */
template<class T>
StringMapElem<T>::StringMapElem(const StringMapElem<T> &other)
  : len(other.len), second(other.second)
{
  first=new char[len+1];
  memcpy(first,other.first,len+1);
}



template<class T>
StringMapElem<T>::StringMapElem(const char *p,int len)
  : len(len)
{
  first=new char[len+1];
  first[len]='\0';
  strncpy(first,p,len);
}



template<class T>
StringMapElem<T>::StringMapElem(const char *p,int len,const T &t)
  : second(t), len(len)
{
  first=new char[len+1];
  first[len]='\0';
  strncpy(first,p,len);
}



template<class T>
StringMapElem<T>::~StringMapElem()
{
  //cout<<"~StringMapElem"<<endl;
  delete [] first;
}



template<class T>
StringMapElem<T> &StringMapIterator<T>::operator*()
{
  return **cur;  
}



template<class T>
StringMapIterator<T> &StringMapIterator<T>::operator++()
{
  if(index<tableSize)
    {
      ++cur;
      if(cur==end) 
	{
	  ++index;
	  findNonemptyBucket();
	}
    }
  return *this;
}



template<class T>
StringMapIterator<T> &StringMapIterator<T>::operator++(int)
{
  if(index<tableSize)
    {
      ++cur;
      if(cur==end) 
	{
	  ++index;
	  findNonemptyBucket();
	}
    }
  return *this;
}



template<class T>
StringMapIterator<T>::StringMapIterator(int index,int tableSize,Bucket *array)
  : array(array), tableSize(tableSize), index(index)
{
  findNonemptyBucket();  
}



template<class T>
bool StringMapIterator<T>::operator!=(const StringMapIterator<T> &i) 
{
  return !(*this==i);
}



template<class T>
bool StringMapIterator<T>::operator==(const StringMapIterator<T> &other)
{
  if(other.index!=index) return false;
  if(index>=tableSize) return true;
  return cur==other.cur;
}



template<class T>
void StringMapIterator<T>::findNonemptyBucket()
{
  for(; index<tableSize ; ++index)
    {
      Bucket &bucket=array[index];
      if(!bucket.empty())
	{
	  cur=bucket.begin();
	  end=bucket.end();
	  return;
	}
    }
}



template<class T>
T &TigrStringMap<T>::lookup(const char *p,int len)
{
  return lookup(p,0,len);
}



template<class T>
T &TigrStringMap<T>::lookup(const char *pOrigin,int index,int len)
{
  const char *p=pOrigin+index;
  unsigned hashValue=hash(p,len,tableSize);
  Bucket &bucket=array[hashValue];
  typename Bucket::iterator end=bucket.end();
  typename Bucket::iterator cur=bucket.begin();
  for(; cur!=end ; ++cur)
    {
      StringMapElem<T> *elem=*cur;
      if(len!=elem->len) continue;
      if(!strncmp(p,elem->first,len))
	return elem->second;
    }
  StringMapElem<T> *newElem=new StringMapElem<T>(p,len);
  bucket.push_back(newElem);
  ++numElements;
  return newElem->second;
}



template<class T>
TigrStringMap<T>::TigrStringMap(const TigrStringMap<T> &other)
  : tableSize(other.tableSize), numElements(other.numElements),
    array(new Bucket[other.tableSize])
{
  for(int i=0 ; i<tableSize ; ++i)
    {
      Bucket &thisBucket=array[i], &thatBucket=other.array[i];
      typename Bucket::iterator cur=thatBucket.begin(), end=thatBucket.end();
      for(; cur!=end ; ++cur)
	thisBucket.push_back(new StringMapElem<T>(**cur));
    }
}



template<class T>
TigrStringMap<T> &TigrStringMap<T>::operator=(const TigrStringMap &other)
{
  tableSize=other.tableSize;
  numElements=other.numElements;
  array=new Bucket[other.tableSize];
  for(int i=0 ; i<tableSize ; ++i)
    {
      Bucket &thisBucket=array[i];
      const Bucket &thatBucket=other.array[i];
      typename Bucket::const_iterator cur=thatBucket.begin(), 
	end=thatBucket.end();
      for(; cur!=end ; ++cur)
	thisBucket.push_back(new StringMapElem<T>(**cur));
    }
  return *this;
}



template<class T>
TigrStringMap<T>::TigrStringMap(int tableSize)
  : tableSize(tableSize), numElements(0)
{
  array=new Bucket[tableSize];
}



template<class T>
typename TigrStringMap<T>::iterator TigrStringMap<T>::begin()
{
  return iterator(0,tableSize,array);
}



template<class T>
typename TigrStringMap<T>::iterator TigrStringMap<T>::end()
{
  return iterator(tableSize,tableSize,array);
}



template<class T>
typename TigrStringMap<T>::iterator TigrStringMap<T>::begin() const
{
  return iterator(0,tableSize,const_cast<Bucket*>(array));
}



template<class T>
typename TigrStringMap<T>::iterator TigrStringMap<T>::end() const
{
  return iterator(tableSize,tableSize,const_cast<Bucket*>(array));
}



template<class T>
TigrStringMap<T>::~TigrStringMap()
{
  //cout<<"~TigrStringMap"<<endl;
  for(int i=0 ; i<tableSize ; ++i)
    {
      Bucket &bucket=array[i];
      typename Bucket::iterator cur=bucket.begin(), end=bucket.end();
      for(; cur!=end ; ++cur)
	delete *cur;
    }
  delete [] array;
}



template<class T>
bool TigrStringMap<T>::isDefined(const char *p,int len)
{
  return isDefined(p,0,len);
}



template<class T>
bool TigrStringMap<T>::isDefined(const char *pOrigin,int index,int len)
{
  const char *p=pOrigin+index;
  unsigned hashValue=hash(p,len,tableSize);
  Bucket &bucket=array[hashValue];
  typename Bucket::iterator cur=bucket.begin(), end=bucket.end();
  for(; cur!=end ; ++cur)
    {
      StringMapElem<T> *elem=*cur;
      if(len!=elem->len) continue;
      if(!strncmp(p,elem->first,len))
	return true;
    }
  return false;
}



template<class T>
int TigrStringMap<T>::size()
{
  return numElements;
}



template<class T>
unsigned TigrStringMap<T>::hash(const char *s,int length,int tableSize)
{
  int h=0;
  const char *p=s, *end=s+length;
  for(; p!=end ; ++p)
    {
      h=(h<<4)+*p;
      int g=h & 0xf000;
      if(g) h=h^(g>>8);
    }
  return labs(h) % tableSize;
}



template<class T>
void TigrStringMap<T>::clear()
{
  for(int i=0 ; i<tableSize ; ++i)
    {
      Bucket &bucket=array[i];
      typename Bucket::iterator cur=bucket.begin(), end=bucket.end();
      for(; cur!=end ; ++cur)
	delete *cur;
      bucket.clear();
    }
  numElements=0;
}



template<class T>
void TigrStringMap<T>::remove(const char *p,int len)
{
  remove(p,0,len);
}



template<class T>
void TigrStringMap<T>::remove(const char *pOrigin,int index,int len)
{
  const char *p=pOrigin+index;
  unsigned hashValue=hash(p,len,tableSize);
  Bucket &bucket=array[hashValue];
  typename Bucket::iterator cur=bucket.begin(), end=bucket.end();
  for(; cur!=end ; ++cur)
    {
      StringMapElem<T> *elem=*cur;
      if(len!=elem->len) continue;
      if(!strncmp(p,elem->first,len))
	{
	  bucket.erase(cur);
	  break;
	}
    }
  --numElements;
}

#endif
